/* CG/skel/cg_06.c */

/****************************************************************
 *                                                              *
 * This file has been written as a sample solution to an        *
 * exercise in a course given at the High Performance           *
 * Computing Centre Stuttgart (HLRS).                           *
 * It is made freely available with the understanding that      *
 * every copy of this file must include this header and that    *
 * HLRS takes no responsibility for the use of the enclosed     *
 * teaching material.                                           *
 *                                                              *
 * This software is partitioned into 16 tasks.                  *
 * Each task can be implemented by one group of the course      *
 * Files: CG/solution/cg_all.c - Solution                       *
 *        CG/skel/cg_01.c      - Skeletons that should be used  *
 *             .. cg_15.c        to implement each task         *
 *        CG/solution/cgp00.c  - Source part of each task       *
 *             ..     cgp16.c                                   *
 *                                                              *
 * Authors: Dr. Rolf Rabenseifner                               *
 *          Felix Triebel                                       *
 *          Michael Speck                                       *
 *          Felix Triebel                                       *
 *          Traugott Streicher                                  *
 *                                                              *
 * Contact: rabenseifner@hlrs.de                                *
 *                                                              *
 ****************************************************************/
                                                           /*EOH*/

/*>>>>>>>>>>>>>>>>>>>>>> begin of task 00 <<<<<<<<<<<<<<<<<<<<<<*/

#include <stdio.h>
#include <stdlib.h>
#ifndef serial
#  include <mpi.h>
#endif
#include <math.h>
#include <malloc.h>
#include <string.h>

/****************** declarations and global data ***************/

/* options: */
int     m,n;         /* size of the physical-heat matrix without borders */
                     /* m=vertical, n=horizontal */
int     iter_max;    /* maximal number of iterations in CG */
double  epsilon;     /* Limit for sqr( || residual vector || ) */
int     print_level; /* print_level = 1: Only result-error and partial result matrix */
                                 /* = 2: and residual norm after each iteration */
                                 /* = 3: and result physical-heat matrix */
                                 /* = 4: and all vector and matrix data in first iteration */
                                 /* = 5: and all vector and matrix data in _all_ iterations */
int     decomp_dims; /* dimensins of the domain decomposition */
int     m_procs, n_procs; /* number of processes in both directions of the 2-dim domain decomposition */
int     mprt, nprt;  /* number of rows and columns when printing the partial result matrix */

/* constants: */

#define MSG_TAG  0
#define MSG_TAG1 1
#define MSG_TAG2 2

#define WITH_HALO    1
#define WITHOUT_HALO 0

/* global structures and data: */

int numprocs, my_rank;

/* start and end+1 of the local chunk */
int start, end1;
int chunk_size;

 
/* data struktures */
struct matrix_row {
        int     size;
        int*    J; /* position in global row */
        int*    J_loc; /* position in local vectors */
        double* value;
};

/* chunk of the A matrix (each process gets one chunk) */
struct matrix_chunk {
        /* size = end1 - start */
        struct matrix_row* row;
} A;


 
/* chunk of a vector */
/* size = end1 - start + halo_size, if halo is used
 *                     + 0        , if halo is not used */
typedef double* vector_chunk;
vector_chunk x, b;

/* halo strukture, etc. */
struct {
    int  size;           /* number of already used elements */
    int  size_allocated; /* number of allocated elements in memory ( >=size ) */
    int* I;              /* used size = halo_info.size, index = halo_pos */
} halo_info;

struct neighbor_struct {
        int     size;    // number of elements to communicate
        int*    I_loc;   // positions in the halo or the real data
        double* scratch; // contiguous scratch array
};

/* neighbors from which we receive data into the halo */
struct neighbor_struct* recv_halo;  // Array of size numprocs
                     // recv_halo[rank].I_loc: positions in the halo, to store the values
int                     recv_num;
int*                    recv_ranks; // Array of size recv_num

/* neighbors which we send original data to */
struct neighbor_struct* send_halo;  // Array of size numprocs
                     // send_halo[rank].I_loc: positions in the real data, to fetch the values
int                     send_num;
int*                    send_ranks; // Array of size recv_num



/********************** memory management **********************/

void abrt(void)
{
#ifndef serial
        MPI_Abort(MPI_COMM_WORLD, -1);
#endif
        exit(-1);
}

/* save_malloc: Same meaning as malloc(); error handling included */
void* save_malloc(int size)
{
        void *new;

        if (size == 0) return NULL;

        new = (void*)malloc(size);
        if (new == NULL) {
                printf("Could not allocate memory.\n");
                abrt();
        } 
        return(new);

}

/*------------------------------
 * This helper function is needed for dynamically growing integer arrays.
 * It takes a pointer to the array and a pointer to a size_allocated variable
 * as arguments.
 *
 * average memory wastage: 33% (in case of normal distribution of size),
 *    worst case 50%
 * average time used for (re)initializing data: 2*sizeof(array_used)
 *    = 1.5 * 2 * sizeof(array_needed) in average case
 *    or 2 * 2 * sizeof(array_needed) in worst case.
 * We use this solution, as it is only needed twice:
 *    for initializing halo and communication info
 *    before computation.
 */
void enlarge_int_array(int** array, int* size_allocated)
{
        /* size_allocated holds the old size of the array and gets altered here. */
        *size_allocated = 2 * (*size_allocated) + 1;  // +1 if size_allocated = 0
        *array = (int*)realloc(*array, sizeof(int) * (*size_allocated));
        if (*array == NULL) {
                printf("Could not allocate memory.\n");
                abrt();
        }
}


/********************** commandline options ********************/

void read_options( int argc, char* argv[] )
{
        int optid = 1; /* start at first optional argument */
        int err_detected = 0;
        if ( my_rank == 0 ) 
        {        
            while ( optid < argc ) {
                if (!strcmp( argv[optid], "-m") ) { optid++;
                    m        = atoi( argv[optid] ); if (m<=0) {err_detected = 1; break;}
                } else if (!strcmp( argv[optid], "-n") ) { optid++;
                    n        = atoi( argv[optid] ); if (n<=0) {err_detected = 1; break;}
                } else if (!strcmp( argv[optid], "-imax") ) { optid++;
                    iter_max = atoi( argv[optid] ); if (iter_max<=0) {err_detected = 1; break;}
                } else if (!strcmp( argv[optid], "-eps") ) { optid++;
                    epsilon  = atof( argv[optid] ); if (epsilon<=0) {err_detected = 1; break;}
                } else if (!strcmp( argv[optid], "-prtlev") ) { optid++;
                    print_level = atoi( argv[optid] ); if (print_level<0) {err_detected = 1; break;}
                } else if (!strcmp( argv[optid], "-twodims") ) { 
                    decomp_dims = 2;
                } else if (!strcmp( argv[optid], "-mprocs") ) { optid++;
                    m_procs = atoi( argv[optid] ); if (m_procs<=0) {err_detected = 1; break;}
                } else if (!strcmp( argv[optid], "-nprocs") ) { optid++;
                    n_procs = atoi( argv[optid] ); if (n_procs<=0) {err_detected = 1; break;}
                } else {err_detected = 1; break;}
                optid++;
            }
            if ( err_detected ) {
                printf( "unknown command line option: %s\n", argv[optid] );
                printf( "\n");
                printf( "Synopsis: mpirun -np <number of processes> ./cg_all <options>\n");
                printf( "  with options [defaults]:\n");
                printf( "  -m <number of rows in physical area of the Laplace equation> [4]\n");
                printf( "  -n <number of columns in physical area of the Laplace equation [4]\n");
                printf( "  -imax <maximal number of iterations> [500]\n");
                printf( "  -eps  <epsion> [1e-6]\n");
                printf( "  -prtlev <print and debug level 0..5> [1]\n");
                printf( "  -twodims [only one-dimensional domain decomposition (DD)]\n");
                printf( "  -mprocs <horizontal number of processes in 2-dim DD\n");
                printf( "  -nprocs <vertical   number of processes in 2-dim DD\n");
                printf( "\n");
                fflush(stdout); abrt();
            }
        }
#ifndef serial
        MPI_Bcast(&m,          1, MPI_INT,    0, MPI_COMM_WORLD);
        MPI_Bcast(&n,          1, MPI_INT,    0, MPI_COMM_WORLD);
        MPI_Bcast(&iter_max,   1, MPI_INT,    0, MPI_COMM_WORLD);
        MPI_Bcast(&epsilon,    1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
        MPI_Bcast(&print_level,1, MPI_INT,    0, MPI_COMM_WORLD);
        MPI_Bcast(&decomp_dims,1, MPI_INT,    0, MPI_COMM_WORLD);
        MPI_Bcast(&m_procs,    1, MPI_INT,    0, MPI_COMM_WORLD);
        MPI_Bcast(&n_procs,    1, MPI_INT,    0, MPI_COMM_WORLD);
#endif
}

/*>>>>>>>>>>>>>>>>>>>>>> -end- of task 00 <<<<<<<<<<<<<<<<<<<<<<*/
/*>>>>>>>>>>>>>>>>>>>>>> begin of task 01 <<<<<<<<<<<<<<<<<<<<<<*/

/*********************** distribution ************************/

/*------------------------------
 * void Distribution
 *  IN  int mat_size : size of the matrix
 *
 * TASK: The heat matrix is distributed in slices of equal size maybe except for
 *       the last process. In this program, the slices are called chunks.
 *       This function calculates the chunk_size, the start and the end1
 *       value for the local chunk. It uses the size of the matrix, the number
 *       of processes and the rank for this calculation.
 *
 * REFERENCE: Slides 13, 15, 24, 26
 *
 * USED GLOBAL DATA:
 *  IN  int my_rank    : rank of the local process (my_rank)
 *  IN  int numprocs   : number of processes
 *  OUT int chunk_size : size of the local chunk
 *  OUT int start      : index for the beginning of the local chunk
 *  OUT int end1       : index+1 for the end of the local chunk
 * USED FUNCTIONS: none
 */
void Distribution(int mat_size)
{
/*01*/  chunk_size=(mat_size-1)/numprocs + 1; // rounding up
/*01*/  start = chunk_size*my_rank;
/*01*/  if (start > mat_size)  start = mat_size;
/*01*/  end1 = chunk_size*(my_rank+1);
/*01*/  if (end1  > mat_size)  end1 = mat_size;
}


/*------------------------------
 * int Row2Rank
 *  IN int     : global_index
 *  RETURNS int: rank
 *
 * TASK: Given a matrix or vector global row index, it calculates
 *       the processor's rank on which it should be located.
 *
 * REFERENCE: see Distribution() above
 *
 * GLOBAL DATA:
 *  IN int chunk_size : size of the local chunk
 * USED FUNCTIONS: none
 */
int Row2Rank(int global_index)
{
/*01*/ return (global_index / chunk_size);
}

/*------------------------------
 * int Rank2StartRow
 *  IN int     : rank
 *  RETURNS int: global start index
 *
 * TASK: Calculates the first global row index for a given rank
 *
 * REFERENCE: see Distribution() above
 *
 * GLOBAL DATA:
 *  IN int chunk_size : size of the local chunk
 * USED FUNCTIONS: none
 */
int Rank2StartRow(int rank)
{
/*01*/  return(chunk_size*rank);
}

/*>>>>>>>>>>>>>>>>>>>>>> -end- of task 01 <<<<<<<<<<<<<<<<<<<<<<*/
/*>>>>>>>>>>>>>>>>>>>>>> begin of task 02 <<<<<<<<<<<<<<<<<<<<<<*/

/********************** vector routines ************************/

/*------------------------------
 * vector_chunk alloc_vec
 *  IN int needs_halo : flag to indicate the need for extra halo memory
 *  RETURNS : address of the vector data array
 *
 * TASK: allocates the part of a vector-data's memory for the calling process
 *       and initializes it with zero.
 *
 * REFERENCE: Slides 16 (see "own data"), 21
 *
 * GLOBAL DATA:
 *  IN int start          : begin of the local chunk
 *  IN int end1           : end+1 of the local chunk
 *  IN int halo_info.size : size of halo data elements
 * USED FUNCTIONS: APPL: save_malloc
 */
vector_chunk alloc_vec(int needs_halo)
{
        int size, I_loc;
        vector_chunk new;
/*02*/  size = (end1-start) + (needs_halo ? halo_info.size : 0);
/*02*/  new = (vector_chunk) save_malloc(sizeof(double) * size);
/*02*/  /* set the vector data to zero */
/*02*/  for(I_loc=0; I_loc < size; I_loc++) new[I_loc]=0.0;
        return new;
}

/*------------------------------
 * void Vec_Set_Value
 *  IN vector_chunk v : set vector v
 *  IN int I          : at global index I
 *  IN double value   : to value
 *
 * TASK: set one value within a vector
 *
 * REFERENCE: Slides 16 (see "own data"), 21
 *
 * GLOBAL DATA:
 *  IN int start   : begin of the local chunk
 *  IN int my_rank : rank of the local process (only for error handling)
 * USED FUNCTIONS: APPL: Row2Rank
 */
void Vec_Set_Value(vector_chunk v, int I, double value)
{
        int I_loc;
/*02*/  if (Row2Rank(I) != my_rank) { printf("ERROR: Vectorelement not local!\n"); abrt(); }
/*02*/  I_loc = I - start;
/*02*/  v[I_loc] = value;
}


/*------------------------------
 * void dupl_vec
 *  IN vector_chunk v1  : vector 1
 *  OUT vector_chunk v2 : vector 2
 *
 * TASK: do a local vector chunk (without halo) duplication: v2 := v1
 *       easy solution: memcpy (see '> man memcpy')
 *
 * REFERENCE: Slides 16 (see "own data"), 21
 *
 * GLOBAL DATA:
 *  IN int start : begin of the local chunk
 *  IN int end1  : end+1 of the local chunk
 * USED FUNCTIONS: C-lib: memcpy
 */
void dupl_vec(vector_chunk v1, vector_chunk v2)
{
/*02*/  memcpy(v2, v1, sizeof(double)*(end1-start));
}

/*------------------------------
 * void add_vec_vec
 *  IN vector_chunk v1  : vector 1
 *  IN double lambda    : factor lambda
 *  IN vector_chunk v2  : vector 2
 *  OUT vector_chunk v3 : vector 3
 *
 * TASK: calculate locally:   v3 := v1 + lambda * v2
 *
 * REFERENCE: Slides 16 (see "own data"), 21
 *
 * GLOBAL DATA:
 *  IN int start : begin of the local chunk
 *  IN int end1  : end+1 of the local chunk
 * USED FUNCTIONS: none
 */
void add_vec_vec(vector_chunk v1, double lambda, vector_chunk v2,
                 vector_chunk /* out */ v3)
{
        int I_loc;

/*02*/  for(I_loc=0; I_loc < (end1-start); I_loc++)
/*02*/          v3[I_loc] = v1[I_loc] + lambda * v2[I_loc];
}

/*------------------------------
 * double dot_product_vec_vec
 *  IN vector_chunk v1 : vector 1
 *  IN vector_chunk v2 : vector 2
 *  RETURNS double     : dot product
 *
 * TASK: calculate global dot product for vector 1 and vector 2
 *
 * REFERENCE: Slides 16 (see "own data"), 21
 *
 * GLOBAL DATA:
 *  IN int start : begin of the local chunk
 *  IN int end1  : end+1 of the local chunk
 * USED FUNCTIONS: APPL: none
 *                 MPI : MPI_Allreduce
 */
double dot_product_vec_vec(vector_chunk v1, vector_chunk v2)
{
        int I_loc;
        double sum=0, global_sum;

        /* dot product of local chunk (without halo) */
/*02*/  for(I_loc=0; I_loc < (end1-start); I_loc++)
/*02*/          sum += v1[I_loc] * v2[I_loc];

        /* global sum */
#ifdef serial
/*02*/  global_sum = sum;
#else
/*02*/  MPI_Allreduce(&sum, &global_sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
#endif

        return global_sum;
}


/*------------------------------
 * double sqr_norm_vec
 *  IN vector_chunk v1 : vector 1
 *  RETURNS double     : square of norm
 *
 * TASK: calculate global (vector norm Euklid)^2 for vector 1
 *
 * REFERENCE: Slides 16 (see "own data"), 21
 *
 * GLOBAL DATA:
 *  IN int start : begin of the local chunk
 *  IN int end1  : end+1 of the local chunk
 * USED FUNCTIONS: APPL: none
 *                 MPI : MPI_Allreduce
 */
double sqr_norm_vec(vector_chunk v1)
{
        int I_loc;
        double sum=0, global_sum;

        /* dot product of local chunk (without halo) */
/*02*/  for(I_loc=0; I_loc < (end1-start); I_loc++)
/*02*/          sum += v1[I_loc] * v1[I_loc];

        /* global sum */
#ifdef serial
/*02*/  global_sum = sum;
#else
/*02*/  MPI_Allreduce(&sum, &global_sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
#endif

        return global_sum;
}

/*------------------------------
 * double max_norm_vec
 *  IN vector_chunk v1 : vector 1
 *  RETURNS double     : maximum norm
 *
 * TASK: calculate global (vector maximum norm) for vector 1
 *
 * REFERENCE: Slides 16 (see "own data"), 21
 *
 * GLOBAL DATA:
 *  IN int start : begin of the local chunk
 *  IN int end1  : end+1 of the local chunk
 * USED FUNCTIONS: APPL: none
 *                 MPI : MPI_Allreduce
 */
double max_norm_vec(vector_chunk v1)
{
        int I_loc;
        double maxabs=0, global_max;

        /* dot product of local chunk (without halo) */
/*02*/  for(I_loc=0; I_loc < (end1-start); I_loc++)
/*02*/  {
/*02*/          if ( v1[I_loc] > maxabs) maxabs =  v1[I_loc];
/*02*/          if (-v1[I_loc] > maxabs) maxabs = -v1[I_loc];
/*02*/  }

        /* global maximum */
#ifdef serial
/*02*/  global_max = maxabs;
#else
/*02*/  MPI_Allreduce(&maxabs, &global_max, 1, MPI_DOUBLE, MPI_MAX, MPI_COMM_WORLD);
#endif

        return global_max;
}

/*>>>>>>>>>>>>>>>>>>>>>> -end- of task 02 <<<<<<<<<<<<<<<<<<<<<<*/

/*>>>>>>>>>>>>>>>>>>>>>> begin of task 03 <<<<<<<<<<<<<<<<<<<<<<*/

/*********************** matrix routines ************************/

/*------------------------------
 * void alloc_A
 *
 * TASK: This function allocates the local part of the matrix A.
 *
 * REFERENCE: Slides 15, 22
 *
 * USED GLOBAL DATA:
 *  IN  int start             : begin of the local chunk
 *  IN  int end1              : end+1 of the local chunk
 *  OUT struct matrix_chunk A : the matrix A
 * USED FUNCTIONS: APPL: save_malloc
 */
void alloc_A(void)
{
        int I_loc; // local row index in matrix (slide 25)

        /* allocate memory for the rows of the matrix */
        /* A.row is an array of size (end1-start) */
/*03*/  A.row = (struct matrix_row*)
/*03*/          save_malloc(sizeof(struct matrix_row) * (end1-start));

        /* for each row set the number of elements to zero */
/*03*/  for(I_loc=0; I_loc < (end1-start); I_loc++) A.row[I_loc].size = 0;
}

/*------------------------------
 * void Alloc_Matrix_Row
 *  IN/OUT struct matrix_chunk* A : heat matrix
 *  IN     int global_row         : global index of the row in the heat matrix
 *  IN     int Number_Of_Elements : number of elements in the row
 *
 * TASK: This function allocates a row for the matrix A.
 *       The number of elements in the row is given.
 *
 * REFERENCE: Slides 15, 22
 *
 * USED GLOBAL DATA:
 *  IN int my_rank : rank of the local process (only for error handling)
 *  IN int start   : start of the local chunk
 * USED FUNCTIONS: APPL: Row2Rank, save_malloc
 */
void Alloc_Matrix_Row(struct matrix_chunk* A, int global_row, int Number_Of_Elements)
{
        struct matrix_row *Row; // local

        /* check if the row is located on this process */
/*03*/  if (Row2Rank(global_row) != my_rank) { printf("ERROR: Matrix row not local!\n"); abrt(); }

        /* allocate */ 
/*03*/  Row = &(A->row[global_row-start]);
/*03*/  Row->value = (double*)save_malloc(sizeof(double)*Number_Of_Elements);
/*03*/  Row->J = (int*)save_malloc(sizeof(int)*Number_Of_Elements);
/*03*/  Row->J_loc = (int*)save_malloc(sizeof(int)*Number_Of_Elements);
}

/*------------------------------
 * void Mat_Set_Value
 *  IN/OUT struct matrix_chunk* A : matrix
 *  IN     int I                  : index I
 *  IN     int J                  : index J
 *  IN     double v               : value
 *
 * TASK: Sets the matrix element located at I,J to the given value.
 *
 * REFERENCE: Slides 15, 22
 *
 * USED GLOBAL DATA:
 *  IN int start : start of the local chunk
 * USED FUNCTIONS: none
 */
void Mat_Set_Value(struct matrix_chunk* A, int I, int J, double v)
{
        struct matrix_row *Row;

/*03*/  Row = &(A->row[I - start]);
/*03*/  Row->value[Row->size] = v;
/*03*/  Row->J[Row->size] = J;
/*03*/  if ( (J>=start) && (J<end1) ) Row->J_loc[Row->size] = J - start;
/*03*/  else                          Row->J_loc[Row->size] = -1; /*undefined*/
/*03*/  (Row->size)++;
}

/*>>>>>>>>>>>>>>>>>>>>>> -end- of task 03 <<<<<<<<<<<<<<<<<<<<<<*/

/*>>>>>>>>>>>>>>>>>>>>>> begin of task 04 <<<<<<<<<<<<<<<<<<<<<<*/

#ifndef serial
/*------------------------------
 * int Add_Element_To_Halo
 *   IN int I:    global index of the element
 *   RETURNS int: position of the element in halo
 *
 * TASK: Add an element to halo_info, if it is not yet existing.
 *       Helper for Init_Vec_Halo(), see below.
 *
 * REFERENCE: Slides 16, 17 
 *
 * USED GLOBAL DATA:
 *  IN OUT struct halo_info
 * USED FUNCTIONS: APPL: enlarge_int_array
 */
int Add_Element_To_Halo(int I)
{
        int halo_pos;
 
        /* We might use the element x for multiple matrix elements */
        /* => Is I already in the halo? If yes, return existing halo_pos */
/*04*/  for (halo_pos=0; halo_pos<halo_info.size; halo_pos++)
/*04*/  {
/*04*/          if (halo_info.I[halo_pos] == I)
/*04*/             return halo_pos;
/*04*/  }
 
        /* Element not found => */
        /*  a) If not enough space, i.e. halo_info.size==halo_info.size_allocated,
               then enlarge array halo_info.I with enlarge_int_array() (see above) */ 
/*04*/  if (halo_info.size_allocated == halo_info.size)
/*04*/          enlarge_int_array(&halo_info.I, &halo_info.size_allocated);
        /*  b) Add it and return new halo_pos */
/*04*/  halo_info.I[halo_info.size] = I;
/*04*/  halo_info.size++;
/*04*/  return (halo_info.size - 1);
}

/*------------------------------
 * void Init_Vec_Halo(void)
 *
 * TASK: Go through A and enter position of needed
 *       non-local vector-elements to halo_info.
 *
 * REFERENCE: Slides 16, 17, 22
 *
 * USED GLOBAL DATA:
 *  IN  int start           : begin of the local chunk
 *  IN  int end1            : end+1 of the local chunk
 *  IN struct matrix_chunk A: Rows contain columns of elements we need.
 *  OUT struct halo_info    : Describes position of non-local elements in the halo
 * USED FUNCTIONS: APPL: Add_Element_To_Halo
 */
void Init_Vec_Halo(void)
{
        int I_loc, j, J;

        /* initialize empty halo_info (i.e., with {0,0,NULL}) */ 
/*04*/  halo_info.size = 0;
/*04*/  halo_info.size_allocated = 0;
/*04*/  halo_info.I = NULL;
        /* compute A.row[I_loc].J_loc[j] by using Add_Element_To_Halo(J) */
/*04*/  for (I_loc=0; I_loc<end1-start; I_loc++)
/*04*/  {
/*04*/          for (j=0; j<A.row[I_loc].size; j++)
/*04*/          {
/*04*/                  J = A.row[I_loc].J[j];
/*04*/                  if ((J<start) || (J>=end1))
/*04*/                          A.row[I_loc].J_loc[j] = Add_Element_To_Halo(J) + end1-start;
/*04*/          }
/*04*/  }
}
#endif

/*>>>>>>>>>>>>>>>>>>>>>> -end- of task 04 <<<<<<<<<<<<<<<<<<<<<<*/

/*>>>>>>>>>>>>>>>>>>>>>> begin of task 05 <<<<<<<<<<<<<<<<<<<<<<*/

#ifndef serial
/*------------------------------
 * void Find_Used_Ranks
 *   IN  struct neighbor_struct* sendrecv_halo : Points to send_halo or recv_halo
 *   OUT int* num                              : Number of communication partners
 *                                               for sendimg or receiving
 *   OUT int** ranks                           : Array of their ranks
 *
 * TASK: This function finds our communication partners' ranks.
 *       Helper for Init_Recv_Info() and Init_Send_Info1(). 
 *
 * REFERENCE: Slide 18
 *
 * USED GLOBAL DATA:
 *   In int numprocs:           Number of MPI processes
 *   IN struct sendrecv_halo:   Describes position of non-local elements in the halo
 * USED FUNCTIONS: APPL: save_malloc
 */
void Find_Used_Ranks(struct neighbor_struct* sendrecv_halo, int* num, int** ranks)
{
        int rank, i, num_neighbors;
        num_neighbors = 0;

        /* calculate the number of neighbors (recv_num, send_num) */
/*05*/  for (rank=0; rank<numprocs; rank++)
/*05*/    if( sendrecv_halo[rank].size > 0)  num_neighbors++;

        /* allocate memory for recv_ranks, send_ranks */
/*05*/  *ranks = (int*)save_malloc(sizeof(int)*num_neighbors);

        /* initialize the array recv_ranks, send_ranks */
/*05*/  i=0;
/*05*/  for (rank=0; rank<numprocs; rank++)
/*05*/    if( sendrecv_halo[rank].size > 0)
/*05*/    {
/*05*/      (*ranks)[i] = rank;
/*05*/      i++;
/*05*/    }

/*05*/  *num = num_neighbors;
}
#endif

#ifndef serial
/*------------------------------
 * void Init_Recv_Info
 *
 * TASK: This function sets up the recv_halo data-structure.
 *
 * REFERENCE: Slides 18, 20, 22
 *
 * USED GLOBAL DATA:
 *   IN struct halo_info:     Describes position of non-local elements in the halo
 *   In int numprocs:                           Number of MPI processes
 *   IN int start:                              start of the local chunk
 *   IN int end1:                               end + 1 of the local chunk
 *   OUT struct neighbor_struct* recv_halo:     Information about data we want to reveive
 *   OUT int recv_num:                          number of processors we receive data from
 *   OUT int* recv_ranks:                       ranks of all real neighbors that are sending halo-data
 * USED FUNCTIONS: APPL: save_malloc, Row2Rank, Find_Used_Ranks
 */
void Init_Recv_Info(void)
{
        int halo_pos, j, rank, size, local_index;

        /* allocate memory for recv_halo */ 
/*05*/  recv_halo = (struct neighbor_struct*)save_malloc(
/*05*/                  sizeof(struct neighbor_struct)*numprocs);

        /* for all processes set recv_halo.size to 0 */
/*05*/  for (rank=0; rank<numprocs; rank++) recv_halo[rank].size = 0;

/*05*/  /* count the elements to communicate for each rank */
/*05*/  for (halo_pos=0; halo_pos < halo_info.size; halo_pos++)
/*05*/  {
/*05*/          rank = Row2Rank(halo_info.I[halo_pos]);
/*05*/          recv_halo[rank].size++;
/*05*/  }

        /* allocate memory for the recv_halo.scratch and recv_halo.I_loc
         * reset recv_halo.size to 0 */
/*05*/  for (rank=0; rank<numprocs; rank++)
/*05*/  {
/*05*/          size=recv_halo[rank].size;
/*05*/          recv_halo[rank].scratch
/*05*/                  = (double*)save_malloc(sizeof(double)*size);
/*05*/          recv_halo[rank].I_loc
/*05*/                  = (int*)save_malloc(sizeof(int)*size);
/*05*/          recv_halo[rank].size = 0;
/*05*/  }


        /* initialize recv_halo.I_loc and recv_halo.size */
/*05*/  for (halo_pos=0; halo_pos < halo_info.size; halo_pos++)
/*05*/  {
/*05*/          local_index = end1-start + halo_pos;
/*05*/          rank = Row2Rank(halo_info.I[halo_pos]);
/*05*/          j = recv_halo[rank].size; /* next free position */
/*05*/          recv_halo[rank].I_loc[j] = local_index;
/*05*/          recv_halo[rank].size++;
/*05*/  }
        /* Use Find_Used_Ranks() to find all neighbors (i.e., ranks) from which halo-data is received */ 
/*05*/  Find_Used_Ranks(recv_halo, &recv_num, &recv_ranks);
}
#endif

/*>>>>>>>>>>>>>>>>>>>>>> -end- of task 05 <<<<<<<<<<<<<<<<<<<<<<*/
/*>>>>>>>>>>>>>>>>>>>>>> begin of task 06 <<<<<<<<<<<<<<<<<<<<<<*/

#ifndef serial
/*------------------------------
 * void Init_Send_Info1(void)
 *
 * TASK: Broadcast number of elements we need from other
 *       processors (and reveive other processors info)
 *
 * REFERENCE: Slides 19, 20, 23
 *
 * USED GLOBAL DATA:
 *  In int numprocs:                     Number of MPI processes
 *  IN neighbor_struct* recv_halo:       Information about data we want to reveive
 *  OUT neighbor_struct* send_halo:      Information about data we have to send
 *  OUT int send_num:                    number of processors we send data to
 *  OUT int* send_ranks:                 ranks of all real neighbors to which halo-data must be sent
 * USED FUNCTIONS: APPL: Find_Used_Ranks
 *                 MPI : MPI_Alltoall
 */
void Init_Send_Info1(void)
{
        int rank;
        int *sendbuf, *recvbuf;

        /* allocate memory for sendbuf, recvbuf, and send_halo */ 






        /* send to all processes number of requested elements */






        /* Use Find_Used_Ranks() to find all neighbors (i.e.,ranks) that have to send halo-data */ 


        /* free scratch memory */ 

}
#endif

/*>>>>>>>>>>>>>>>>>>>>>> -end- of task 06 <<<<<<<<<<<<<<<<<<<<<<*/

/*>>>>>>>>>>>>>>>>>>>>>> begin of task 07 <<<<<<<<<<<<<<<<<<<<<<*/

#ifndef serial
/*------------------------------
 * void Init_Send_Info2(void)
 *
 * TASK: Communicate positions of elements needed from
 *       (or by) other processors.
 *
 * REFERENCE: Slides 19, 20, 23
 *
 * USED GLOBAL DATA:
 *   IN struct halo_info:     Describes position of non-local elements in the halo
 *   IN int recv_num:                      number of processors we receive data from
 *   IN int* recv_ranks:                   ranks of all real neighbors that are sending halo-data
 *   IN neighbor_struct* recv_halo:        Information about data we want to reveive
 *   IN int send_num:                      number of processors we send data to
 *   IN int* send_ranks:                   ranks of all real neighbors to which halo-data must be sent
 *   IN int start:                         begin of the local chunk
 *   IN OUT neighbor_struct* send_halo:    Information about data we have to send
 *                                         (IN send_halo.size)
 * USED FUNCTIONS: APPL: save_malloc
 *                 MPI : MPI_Irecv, MPI_Send, MPI_Waitall
 */
void Init_Send_Info2(void)
{
        int i, rank, j, local_index, index_in_halo_info;
        MPI_Request *req;
        MPI_Status *statuses;
        int src_rank, dest_rank;
        int *orig_pos;
 
        /* allocate scratch memory for req and statuses */ 
/*07*/  req = (MPI_Request*)save_malloc(sizeof(MPI_Request)*numprocs);
/*07*/  statuses = (MPI_Status*)save_malloc(sizeof(MPI_Status)*numprocs);

        /* allocate memory (send_halo[rank].size elements) for send_halo[...].I_loc and .scratch */ 
/*07*/  for (i=0; i<send_num; i++)
/*07*/      {
/*07*/          rank = send_ranks[i];
/*07*/          send_halo[rank].scratch = (double*)save_malloc(sizeof(double)*send_halo[rank].size);
/*07*/          send_halo[rank].I_loc = (int*)save_malloc(sizeof(int)*send_halo[rank].size);
/*07*/      }

        /* Communicate halo_info */
        /* non-blocking: receive from all processes halo info */
/*07*/  for (i=0; i<send_num; i++)
/*07*/      {
/*07*/          src_rank = send_ranks[i];
/*07*/          MPI_Irecv(send_halo[src_rank].I_loc,
/*07*/                          send_halo[src_rank].size, MPI_INT, src_rank,
/*07*/                          MSG_TAG1, MPI_COMM_WORLD, &(req[i]));
/*07*/      }
 
        /* blocking: send */
/*07*/  for (i=0; i<recv_num; i++)
/*07*/      {
/*07*/          dest_rank = recv_ranks[i];
/*07*/          orig_pos = (int*)save_malloc(sizeof(int)*recv_halo[dest_rank].size);
/*07*/          for (j=0; j < recv_halo[dest_rank].size; j++)
/*07*/          {
/*07*/                  local_index = recv_halo[dest_rank].I_loc[j];
/*07*/                  index_in_halo_info = local_index - (end1-start);
/*07*/                  orig_pos[j] = halo_info.I[index_in_halo_info];
/*07*/                  // sending the global index;
/*07*/                  // local index is computed in the loop below: ... = ... - start
/*07*/          }
/*07*/          MPI_Send(orig_pos,
/*07*/                     recv_halo[dest_rank].size, MPI_INT,
/*07*/                     dest_rank, MSG_TAG1, MPI_COMM_WORLD);
/*07*/          free(orig_pos);
/*07*/      }
 
        /* Waitall */
/*07*/  if (send_num > 0)
/*07*/          MPI_Waitall(send_num, req, statuses);

        /* update send_halo[_].I_loc[_] */
/*07*/  for (i=0; i<send_num; i++)
/*07*/      {
/*07*/          src_rank = send_ranks[i];
/*07*/          for (j=0; j < send_halo[src_rank].size; j++)
/*07*/          {
/*07*/                  send_halo[src_rank].I_loc[j]
/*07*/                           = send_halo[src_rank].I_loc[j] - start;
/*07*/          }
/*07*/      }
        /* free scratch memory */ 
/*07*/  free(req); free(statuses);

}
#endif

/*>>>>>>>>>>>>>>>>>>>>>> -end- of task 07 <<<<<<<<<<<<<<<<<<<<<<*/

/*>>>>>>>>>>>>>>>>>>>>>> begin of task 08 <<<<<<<<<<<<<<<<<<<<<<*/

#ifndef serial
/*------------------------------
 * void comm_vec_halo
 *
 * IN OUT vector_chunk v1
 *
 * TASK: Communicate halo of a vector: send requested elemets
 *       to others processors and update local halo.
 *
 * REFERENCE: Slides 18, 19, 20, 23
 *
 * USED GLOBAL DATA:
 *   IN int recv_num:                      number of processors we receive data from
 *   IN int* recv_ranks:                   ranks of all real neighbors that are sending halo-data
 *   IN (OUT) neighbor_struct* recv_halo:  (recv_halo[all].scratch modified)
 *   IN int send_num:                      number of processors we send data to
 *   IN int* send_ranks:                   ranks of all real neighbors to which halo-data must be sent
 *   IN (OUT) neighbor_struct* send_halo:  (send_halo[all].scratch modified)
 * USED FUNCTIONS: APPL: save_malloc
 *                 MPI : MPI_Irecv, MPI_ISend, MPI_Waitall
 */
void comm_vec_halo(vector_chunk v1)
{
        int i, j, pos, data_size, src_rank, dest_rank;
        MPI_Request *req;
        MPI_Status *statuses;

        /* allocate scratch memory for req and statuses */ 
/*08*/  req = (MPI_Request*)save_malloc(sizeof(MPI_Request)*(recv_num+send_num));
/*08*/  statuses= (MPI_Status*)save_malloc(sizeof(MPI_Status)*(recv_num+send_num));

        /* receive values from neighbors */
/*08*/  for (i=0; i<recv_num; i++)
/*08*/  {
/*08*/          src_rank = recv_ranks[i];
/*08*/          data_size = recv_halo[src_rank].size;
/*08*/          /* receive data non-blocking */
/*08*/          MPI_Irecv(recv_halo[src_rank].scratch, data_size,
/*08*/                   MPI_DOUBLE, src_rank, MSG_TAG2,
/*08*/                   MPI_COMM_WORLD, &(req[i]));
/*08*/  }

        /* send values to neighbors */
/*08*/  for (i=0; i<send_num; i++)
/*08*/  {
/*08*/          dest_rank = send_ranks[i];
/*08*/          data_size = send_halo[dest_rank].size;
/*08*/          /* fill send_halo[for all].scratch[] */
/*08*/          for (j=0; j<data_size; j++)
/*08*/          {
/*08*/                  pos = send_halo[dest_rank].I_loc[j];
/*08*/                  send_halo[dest_rank].scratch[j] = v1[pos];
/*08*/          }
/*08*/          /* send data non-blocking */
/*08*/          MPI_Isend(send_halo[dest_rank].scratch, data_size,
/*08*/                     MPI_DOUBLE, dest_rank, MSG_TAG2,
/*08*/                     MPI_COMM_WORLD, &(req[recv_num+i]));
/*08*/                                      //   ^__ use second part of req array
/*08*/  }

        /* Waitall */
/*08*/  if (recv_num+send_num > 0)
/*08*/          MPI_Waitall(recv_num+send_num, req, statuses);

        /* receive values from neighbors */
/*08*/  for (i=0; i<recv_num; i++)
/*08*/  {
/*08*/          src_rank = recv_ranks[i];
/*08*/          data_size = recv_halo[src_rank].size;
/*08*/          for (j=0; j<data_size; j++)
/*08*/          {
/*08*/                  pos = recv_halo[src_rank].I_loc[j];
/*08*/                  v1[pos] = recv_halo[src_rank].scratch[j];
/*08*/          }
/*08*/  }
 
        /* free scratch memory */ 
/*08*/  free(req); free(statuses);
}
#endif

/*>>>>>>>>>>>>>>>>>>>>>> -end- of task 08 <<<<<<<<<<<<<<<<<<<<<<*/

/*>>>>>>>>>>>>>>>>>>>>>> begin of task 09 <<<<<<<<<<<<<<<<<<<<<<*/

/*------------------------------
 * double dot_product_row_vec
 * IN struct matrix_row* r:    Row of a matrix
 * IN vector_chunk v1:         Vector with halo
 *
 * TASK: Calculates dot product {row of matrix} * {vector_with_halo}.
 *       Helper for mult_A_vec()
 *
 * REFERENCE: Slides 14, 15, 16, 23
 *
 * USED GLOBAL DATA:
 * USED FUNCTIONS: none
 */
double dot_product_row_vec(struct matrix_row* r, vector_chunk v1)
{
        int j, J_loc;
        double sum=0;

        /* dot product of local chunk */
/*09*/  for(j=0; j < r->size; j++)
/*09*/  {
/*09*/          /* get position */
/*09*/          J_loc=r->J_loc[j];
/*09*/          sum += r->value[j] * v1[J_loc];
/*09*/  }

/*09*/  return sum;
}

/*------------------------------
 * mult_A_vec
 *      IN  v1: vector_chunk v1
 *      OUT v2: vector_chunk
 *
 * TASK: Calculates dot product {row of matrix} * {vector}
 *       for all rows of the local chunk.
 *
 * REFERENCE: Slides 14, 15, 16, 23
 *
 * USED GLOBAL DATA: 
 *  IN start, end1
 * USED FUNCTIONS: APPL: dot_product_row_vec
 */
void mult_A_vec(vector_chunk v1, vector_chunk /* out */ v2)
{
        int I_loc;

/*09*/  for(I_loc=0; I_loc < (end1-start); I_loc++)
/*09*/          v2[I_loc] = dot_product_row_vec(&(A.row[I_loc]), v1);
}

/*>>>>>>>>>>>>>>>>>>>>>> -end- of task 09 <<<<<<<<<<<<<<<<<<<<<<*/
/*>>>>>>>>>>>>>>>>>>>>>> begin of task 10 <<<<<<<<<<<<<<<<<<<<<<*/

/******************* domain decomposition **********************/

/*----------------------- 1-dim decomposition -----------------*/

//   Example: m=5, n=7
//
//       |       j=0     1     2     3     4     5     6 |
//   ----+-----------------------------------------------+
//       |      I=                                       |
//   i=0 |         0     1     2     3     4     5     6 |
//     1 |         7     8     9    10    11    12    13 |
//     2 |        14    15    16    17    18    19    20 |
//     3 |        21    22    23    24    25    26    27 |
//     4 |        28    29    30    31    32    33    34 |
//   ----+-----------------------------------------------+

/*------------------------------
 * Init_1dim
 *
 * TASK: None (except printing).
 *
 * REFERENCE: Slides 6, 26
 *
 * USED GLOBAL DATA: none
 * USED FUNCTIONS: none
 */
void Init_1dim(void)
{
        if ((print_level >= 1) && (my_rank == 0))
            printf("Using 1-dim domain decomposition, numprocs=%d\n", numprocs);
}


/*------------------------------
 * I2ij_1dim
 *      IN  I:  global matrix row index
 *      OUT i:  vertical   index of the physical Laplace problem
 *      OUT j:  horizontal index of the physical Laplace problem
 *
 * TASK: Calculates i and j according the scheme in the example above.
 *
 * REFERENCE: Slides 6, 26
 *
 * USED GLOBAL DATA: 
 *  IN int n: vertical size of the physical Laplace problem
 * USED FUNCTIONS: none
 */
void I2ij_1dim(int I, int* i, int* j)
{
/*10*/  (*i) = I/n; (*j) = I - (*i)*n;
}

/*------------------------------
 * ij2I_1dim
 *      IN  i:  vertical   index of the physical Laplace problem
 *      IN  j:  horizontal index of the physical Laplace problem
 *      OUT I:  global matrix row index
 *
 * TASK: Calculates I according the scheme in the example above.
 *
 * REFERENCE: Slides 6, 26
 *
 * USED GLOBAL DATA: 
 *  IN int n: vertical size of the physical Laplace problem
 * USED FUNCTIONS: none
 */
void ij2I_1dim(int i, int j, int* I)
{
/*10*/  (*I) = i*n + j;
}

/*----------------------- 2-dim decomposition -----------------*/

//   Example: m=5, n=7, m_procs=2, n_procs=2
//
//       |       j=0     1     2     3  |  4     5     6 |
//   ----+------------------------------+----------------+  v
//       |      I=       (jj=0)               (jj=1)     |
//   i=0 |         0     1     2     3  | 12    13    14 |  i_
//       |                                               |  chunk_
//     1 | (ii=0)  4     5     6     7  | 15    16    17 |  size
//       |                                               |  = 3
//     2 |         8     9    10    11  | 18    19    20 |
//   - - + - -  - - - - - - - - - - - - + - - - - - - - -+  ^
//     3 |        21    22    23    24  | 29    30    31 |
//       | (ii=1)                                        |
//     4 |        25    26    27    28  | 32    33    34 |
//   ----+------------------------------+----------------+
//
//       >    j_chunk_size = 4     <
//
// Caution: Only if the m and n are multiples of _proc and m_proc
//          then the default distibution fits to the domain
//          decomposition. In this example it does not fit:
//          Process 0 works on I =  0.. 8
//          Process 1 works on I =  9..17
//          Process 2 works on I = 18..26
//          Process 3 works on I = 27..34

int i_chunk_size, j_chunk_size, i_n_size;

/*------------------------------
 * Init_2dim
 *
 * TASK: None (except printing).
 *
 * REFERENCE: Slides 26, 24
 *
 * USED GLOBAL DATA: 
 *  OUT int i_chunk_size, j_chunk_size: see figure above
 *  OUT int i_n_size := i_chunk_size * n
 * USED FUNCTIONS: none
 */
void Init_2dim(void)
{
        if ((print_level >= 1) && (my_rank == 0))
            printf("Using 2-dim domain decomposition, m_procs=%d, n_procs=%d, numprocs=%d\n",
                    m_procs, n_procs, numprocs);
/*10*/  i_chunk_size = (m-1)/m_procs + 1; /* rounded up*/
/*10*/  i_n_size = i_chunk_size * n;
/*10*/  j_chunk_size = (n-1)/n_procs + 1; /* rounded up*/
}


/*------------------------------
 * I2ij_2dim
 *      IN  I:  global matrix row index
 *      OUT i:  vertical   index of the physical Laplace problem
 *      OUT j:  horizontal index of the physical Laplace problem
 *
 * TASK: Calculates i and j according the scheme in the example above.
 *
 * REFERENCE: Slides 26, 24
 *
 * USED GLOBAL DATA: 
 *  IN int i_chunk_size, j_chunk_size: see figure above
 *  IN int i_n_size := i_chunk_size * n
 *  IN int m: horizontal size of the physical Laplace problem
 *  IN int n: vertical   size of the physical Laplace problem
 * USED FUNCTIONS: none
 */
void I2ij_2dim(int I, int* i_arg, int* j_arg)
{
        int i, j, ii, jj, m_slice, n_slice, i_j_size, I_tail;

/*10*/  i = 0;
/*10*/  ii = I / i_n_size;
/*10*/  i = i + ii * i_chunk_size;
/*10*/  m_slice = m - ii * i_chunk_size;
/*10*/  if (m_slice > i_chunk_size) m_slice = i_chunk_size;
/*10*/  I_tail = I - ii * i_n_size;

/*10*/  j = 0;
/*10*/  i_j_size = m_slice * j_chunk_size;
/*10*/  jj = I_tail / i_j_size;
/*10*/  j = j + jj * j_chunk_size;
/*10*/  I_tail = I_tail - jj * i_j_size;
/*10*/  n_slice = n - jj * j_chunk_size;
/*10*/  if (n_slice > j_chunk_size) n_slice = j_chunk_size;

/*10*/  i = i + I_tail / n_slice;
/*10*/  j = j + I_tail % n_slice;

/*10*/  (*i_arg) = i;   (*j_arg) = j;
}

/*------------------------------
 * ij2I_2dim
 *      IN  i:  vertical   index of the physical Laplace problem
 *      IN  j:  horizontal index of the physical Laplace problem
 *      OUT I:  global matrix row index
 *
 * TASK: Calculates I according the scheme in the example above.
 *
 * REFERENCE: Slides 26, 24
 *
 * USED GLOBAL DATA: 
 *  IN int i_chunk_size, j_chunk_size: see figure above
 *  IN int i_n_size := i_chunk_size * n
 *  IN int m: horizontal size of the physical Laplace problem
 *  IN int n: vertical   size of the physical Laplace problem
 * USED FUNCTIONS: none
 */
void ij2I_2dim(int i, int j, int* I)
{
        int ii, jj, i_tail, j_tail, m_slice, n_slice;

/*10*/  ii     = i / i_chunk_size;    jj     = j / j_chunk_size;
/*10*/  i_tail = i % i_chunk_size;    j_tail = j % j_chunk_size;
/*10*/  m_slice = m - ii * i_chunk_size;
/*10*/  if (m_slice > i_chunk_size) m_slice = i_chunk_size;
/*10*/  n_slice = n - jj * j_chunk_size;
/*10*/  if (n_slice > j_chunk_size) n_slice = j_chunk_size;
/*10*/  (*I) = ii*i_n_size + jj*(m_slice * j_chunk_size) + i_tail*n_slice + j_tail;
}

/*-------- switching between 1-dim and 2-dim decomposition -------*/
void Init_Decomp(void)                    /* description, see Init_1dim() above */
{       switch (decomp_dims)
        {       case 1: Init_1dim(); break;
                case 2: Init_2dim(); break;
                default: printf("Wrong decomp_dims=%d\n",decomp_dims); abrt();
        }
}
void I2ij(int I, int* i_arg, int* j_arg)  /* description, see I2ij_1dim() above */
{       switch (decomp_dims)
        {       case 1: I2ij_1dim(I,i_arg,j_arg); break;
                case 2: I2ij_2dim(I,i_arg,j_arg); break;
                default: printf("Wrong decomp_dims=%d\n",decomp_dims); abrt();
        }
}
void ij2I(int i, int j, int* I)           /* description, see ij2I_1dim() above */
{       switch (decomp_dims)
        {       case 1: ij2I_1dim(i,j,I); break;
                case 2: ij2I_2dim(i,j,I); break;
                default: printf("Wrong decomp_dims=%d\n",decomp_dims); abrt();
        }
}
/*>>>>>>>>>>>>>>>>>>>>>> -end- of task 10 <<<<<<<<<<<<<<<<<<<<<<*/

/*>>>>>>>>>>>>>>>>>>>>>> begin of task 11 <<<<<<<<<<<<<<<<<<<<<<*/

/************************ application *************************/

/*------------------------------
 * void Init_Laplace_A
 *  IN/OUT struct matrix_chunk * A : the local chunk of matrix A
 *
 * TASK: go through the local chunk of matrix A,
 *       allocate each matrix row and set it's values
 *
 * REFERENCE: Slides 28, 24
 *
 * GLOBAL DATA:
 *  IN int start : begin of the local chunk
 *  IN int end1  : end+1 of the local chunk
 * USED FUNCTIONS: APPL: Alloc_Matrix_Row, I2ij, ij2I, Mat_Set_Value
 */
void Init_Laplace_A(struct matrix_chunk *A)
{
        int I, J, i, j;

/*11*/  for (I=start; I<end1; I++) {
/*11*/          Alloc_Matrix_Row(A, I, 5 /* not more than 5 values on next lines*/);
/*11*/          I2ij(I, &i, &j);
/*11*/          if (i>0)   {ij2I(i-1,j,&J); Mat_Set_Value(A,I,J,-1.0);}
/*11*/          if (j>0)   {ij2I(i,j-1,&J); Mat_Set_Value(A,I,J,-1.0);}
/*11*/          J = I;                      Mat_Set_Value(A,I,J, 4.0);
/*11*/          if (j<n-1) {ij2I(i,j+1,&J); Mat_Set_Value(A,I,J,-1.0);}
/*11*/          if (i<m-1) {ij2I(i+1,j,&J); Mat_Set_Value(A,I,J,-1.0);}
/*11*/  }
}

/*>>>>>>>>>>>>>>>>>>>>>> -end- of task 11 <<<<<<<<<<<<<<<<<<<<<<*/
/*>>>>>>>>>>>>>>>>>>>>>> begin of task 12 <<<<<<<<<<<<<<<<<<<<<<*/

/*------------------------------
 * void Init_Boundary_Vec
 *  OUT vector_chunk b: the local chunk of vector b
 *
 * TASK: Store all boundary values in the vector memory 
 *       (that is already allocated).
 *       In this example, the boundaries are given by:
 *         u(i,j) = h * (i+1)  with  h = 1.0/(m+1) for i=-1,m and j=-1,n 
 *
 * REFERENCE: Slides 28, 24
 *
 * GLOBAL DATA:
 *  IN int start : begin of the local chunk
 *  IN int end1  : end+1 of the local chunk
 * USED FUNCTIONS: APPL: I2ij, Vec_Set_Value
 */
void Init_Boundary_Vec(vector_chunk b)
{
        int I, i, j;   double v, h;
 
/*12*/  h = 1.0/(m+1);
/*12*/  for (I=start; I<end1; I++) {
/*12*/          I2ij(I, &i, &j);
/*12*/          v = 0.0;
/*12*/          if (i==0)   v = v + /*u(-1,j):*/  h * 0;     /* = 0 */
/*12*/          if (i==m-1) v = v + /*u(m, j):*/  h * (m+1); /* = 1 */
/*12*/          if (j==0)   v = v + /*u(i,-1):*/  h * (i+1);
/*12*/          if (j==n-1) v = v + /*u(i, n):*/  h * (i+1);
/*12*/          if (v != 0) Vec_Set_Value(b,I,v);
/*12*/  }
}

/*------------------------------
 * void Init_Exact_Solution_Vec
 *  OUT vector_chunk u: the local chunk of vector u
 *
 * TASK: Store all boundary values in the vector memory 
 *       (that is already allocated).
 *       In this example, the inner values are given by:
 *       u(i,j) = h * (i+1)  with  h = 1.0/(m+1) for i=0..m-1 and j=0..n-1 
 *
 * REFERENCE: Slides 28, 24
 *
 * GLOBAL DATA:
 *  IN int start : begin of the local chunk
 *  IN int end1  : end+1 of the local chunk
 * USED FUNCTIONS: APPL: I2ij, Vec_Set_Value
 */
void Init_Exact_Solution_vec(vector_chunk u)
{
        int I, i, j;   double v, h;
 
/*12*/  h = 1.0/(m+1);
/*12*/  for (I=start; I<end1; I++) {
/*12*/          I2ij(I, &i, &j);
/*12*/          v = /*u(i, j):*/ h * (i+1); Vec_Set_Value(u,I,v);
/*12*/  }
}
/*>>>>>>>>>>>>>>>>>>>>>> -end- of task 12 <<<<<<<<<<<<<<<<<<<<<<*/

/*>>>>>>>>>>>>>>>>>>>>>> begin of task 13 <<<<<<<<<<<<<<<<<<<<<<*/

void appl_print_vec(char *name, double *v, int mprint, int nprint)
{
        int i,j, I;
        if (my_rank == 0)
        {
                printf("%s:       j=",name);
                    for (j=0; j<n; j++)
                            if ((n>nprint ? (nprint-1)*(j+n-2)/(n-1) != (nprint-1)*(j+n-1)/(n-1) : 1))
                                printf(" %5d",j); printf("\n");

                for (i=0; i<m; i++)
                  if ((m>mprint ? (mprint-1)*(i+m-2)/(m-1) != (mprint-1)*(i+m-1)/(m-1) : 1))
                  {
                    printf("%s: i=%3d   ",name,i);
                    for (j=0; j<n; j++)
                    {
                        ij2I(i, j, &I);
                        if ((n>nprint ? (nprint-1)*(j+n-2)/(n-1) != (nprint-1)*(j+n-1)/(n-1) : 1))
                            printf(" %5.3f",v[I]);
                    }
                    printf("\n");
                  }
        }
}

/*>>>>>>>>>>>>>>>>>>>>>> -end- of task 13 <<<<<<<<<<<<<<<<<<<<<<*/
/*>>>>>>>>>>>>>>>>>>>>>> begin of task 14 <<<<<<<<<<<<<<<<<<<<<<*/

/************************ print results ************************/

void recv_token(void)
{
#ifndef serial
        int dummy; 
        MPI_Status status;
        if (my_rank > 0)
                MPI_Recv(&dummy, 0, MPI_INT, my_rank-1, MSG_TAG, MPI_COMM_WORLD, &status);
#endif
}

void flush_and_send_token(void)
{
#ifndef serial
        MPI_Status status;
        int dummy; 
#endif
        fflush(stdout);
#ifndef serial
        MPI_Send(&dummy, 0, MPI_INT, (my_rank+1) % numprocs, MSG_TAG, MPI_COMM_WORLD);
        /* Process 0must wait until the last process (numprocs-1) has finished its output */ 
        if (my_rank == 0)
                MPI_Recv(&dummy, 0, MPI_INT, numprocs-1, MSG_TAG, MPI_COMM_WORLD, &status);
#endif
}

void Print_Decomp(void)
{
        int i,j, I, inew,jnew;
        if (my_rank == 0)
        {
                for (i=0; i<m; i++)
                {
                    printf("Decomp: i=%2d",i);
                    for (j=0; j<n; j++)
                    {
                        ij2I(i, j, &I);
                        I2ij(I, &inew, &jnew);
                        printf(" I(%1d,%1d)=%2d%c",inew,jnew,I, ((i==inew)&&(j==jnew) ? ' ' : '!') );
                    }
                    printf("\n");
                }
        }
        fflush(stdout);
}

void print_vec(char *name, vector_chunk v1)
{
        int I;
        recv_token();
        for (I=start; I<end1; I++)
                printf("%s[%2d] = %6.3f\n", name, I, v1[I-start]);
        flush_and_send_token();
}

void print_vec2d(char *name, vector_chunk v2, int mprint, int nprint)
{
        int I, rank;
        double* whole_v2;
#ifndef serial
        MPI_Status status;
#endif


        whole_v2 = (double*)save_malloc(sizeof(double)*n*m);
#ifndef serial
        if (my_rank > 0)
        {
                MPI_Send(v2, end1-start, MPI_DOUBLE, 0, MSG_TAG, MPI_COMM_WORLD);
        } else {
                for (rank=1; rank<numprocs; rank++)
                {
                        MPI_Recv(&(whole_v2[Rank2StartRow(rank)]), n*m, MPI_DOUBLE, 
                                 rank, MSG_TAG, MPI_COMM_WORLD, &status);
                }
        }
#endif
        for (I=start; I<end1; I++) whole_v2[I] = v2[I-start];
        appl_print_vec(name, whole_v2, mprint, nprint);
        free(whole_v2);
}

void print_mat(char *name, struct matrix_chunk *A)
{
        int I, k;
        recv_token();
        for (I=start; I<end1; I++)
        {
                printf("%s[%2d] = ", name, I);
                for (k=0; k<A->row[I-start].size; k++)
                        printf(" %1d(%1d):%2.0f", A->row[I-start].J[k], A->row[I-start].J_loc[k],
                                        A->row[I-start].value[k]);
                printf("\n");
        }
        flush_and_send_token();
}

void print_halo(void)
{
        int I_loc, halo_pos;
        recv_token();
        printf("Halo: my_rank =%3d - local index:", my_rank);
          for (I_loc=0; I_loc<end1-start;     I_loc++) printf(" %2d",I_loc);printf(" |");
          for (halo_pos=0; halo_pos<halo_info.size; halo_pos++) printf(" %2d",end1-start+halo_pos);printf("\n");

        printf("                   - orig. index:");
          for (I_loc=0; I_loc<end1-start;     I_loc++) printf(" %2d",start+I_loc);printf(" |");
          for (halo_pos=0; halo_pos<halo_info.size; halo_pos++) printf(" %2d",halo_info.I[halo_pos]);
          printf("\n");
        printf("                   - size=%d,  size_allocated=%d\n",
                          halo_info.size, halo_info.size_allocated);
        flush_and_send_token();
}

void print_vec_withhalo(char *name, vector_chunk v1)
{
        int I_loc, halo_pos;
        recv_token();
        printf("Halo: my_rank =%3d - local index:", my_rank);
          for (I_loc=0; I_loc<end1-start;     I_loc++) printf(" %3d",I_loc);printf(" |");
          for (halo_pos=0; halo_pos<halo_info.size; halo_pos++) printf(" %3d",end1-start+halo_pos);
          printf("\n");

        printf("                   - orig. index:");
          for (I_loc=0; I_loc<end1-start;     I_loc++) printf(" %3d",start+I_loc);printf(" |");
          for (halo_pos=0; halo_pos<halo_info.size; halo_pos++) printf(" %3d",halo_info.I[halo_pos]);
          printf("\n");
        printf("                   - vector %s:   ",name);
          for (I_loc=0; I_loc<end1-start; I_loc++) printf(" %3.1f",v1[I_loc]);printf(" |");
          for (I_loc=end1-start; I_loc<end1-start+halo_info.size; I_loc++) printf(" %3.1f",v1[I_loc]);
          printf("\n");
        printf("                   - size=%d,  size_allocated=%d\n",
                          halo_info.size, halo_info.size_allocated);
        flush_and_send_token();
}

void print_recv_info(void)
{
        int i, rank;
        recv_token();
        printf("Recv: my_rank =%3d - recv_num = %d, recv_ranks:",
                my_rank, recv_num);
        for (i=0; i<recv_num; i++) printf(" %2d",recv_ranks[i]);printf("\n");
        for (rank=0; rank<numprocs; rank++)
        {       printf("                   - recv_halo[%2d].size=%2d,",rank,recv_halo[rank].size);
                printf(" .I_loc:");
                for (i=0; i<recv_halo[rank].size; i++)
                        printf(" %2d",recv_halo[rank].I_loc[i]);
                printf("\n");
        }
        flush_and_send_token();
}


void print_send_info(void)
{
        int i, rank;
        recv_token();
        printf("Send: my_rank =%3d - send_num = %d, send_ranks:",
                my_rank, send_num);
        for (i=0; i<send_num; i++) printf(" %2d",send_ranks[i]);printf("\n");
        for (rank=0; rank<numprocs; rank++)
        {       printf("                   - send_halo[%2d].size=%2d,",rank,send_halo[rank].size);
                printf(" .I_loc:");
                for (i=0; i<send_halo[rank].size; i++)
                        printf(" %2d",send_halo[rank].I_loc[i]);
                printf("\n");
        }
        flush_and_send_token();
}

/*>>>>>>>>>>>>>>>>>>>>>> -end- of task 14 <<<<<<<<<<<<<<<<<<<<<<*/
/*>>>>>>>>>>>>>>>>>>>>>> begin of task 15 <<<<<<<<<<<<<<<<<<<<<<*/

/************************ CG solver ****************************/

/*------------------------------
 * void CG_solve
 * OUT int *num_iter  : return the number of iterations
 * OUT double *norm_r : return the norm of r
 *
 * TASK: solve Laplace equation
 *
 * REFERENCE: Slide 12, 24
 *
 * GLOBAL DATA:
 * IN int iter_max       : maximal number of iterations
 * IN double epsilon     : epsilon for maximal error
 * IN struct matrix_chunk A: Rows contain columns of elements we need.
 * IN vector_chunk b     : local chunk of vector b
 * IN/OUT vector_chunk x : local chunk of vector x
 * USED FUNCTIONS: APPL: alloc_vec, dupl_vec, comm_vec_halo, mult_A_vec,
 *                       add_vec_vec, sqr_norm_vec, dot_product_vec_vec
 */
void CG_solve(int *num_iter, double *norm_r)
{
        /* only p has a halo */
        vector_chunk p, r, v;
        double alpha, alpha_new, lambda;
        int iter;

        /* allocate vectors p, r, v */ 
/*15*/  p = alloc_vec(1);
/*15*/  r = alloc_vec(0);
/*15*/  v = alloc_vec(0);

        /* x is initialized with 0 */

        /* instead of p = Ax */
           /* p := x */
/*15*/     dupl_vec(x, p);
#ifndef serial
           /* update halo */
/*15*/     comm_vec_halo(p);
#endif

           /* v = Ap */
/*15*/     mult_A_vec(p, v);
           if (print_level>=4) print_vec("v", v);

           /* p = b - v */
/*15*/     add_vec_vec(b, -1, v, p);
           if (print_level>=4) print_vec("p", p);

        /* r := p */
/*15*/  dupl_vec(p, r);
        if (print_level>=4) print_vec("r", r);

        /* alpha = ||r||**2,   Caution: sqr_norm_vec() computes already ||.||**2 */
/*15*/  alpha = sqr_norm_vec(r);


        for(iter=0; (iter < iter_max) && (alpha > epsilon) ; iter++)
        {
                if ((print_level>=5) || ( (print_level>=4)&&(iter==0) )) {
                printf("CG - my_rank =%3d, i=%d, alpha=%9.6f\n",my_rank,iter,alpha); fflush(stdout);
                }

#ifndef serial
                /* update halo */
/*15*/          comm_vec_halo(p);
#endif

                if ((print_level>=5) || ( (print_level>=4)&&(iter==0) )) {
                        print_mat("A", &A);
                        print_vec("p", p);
                        print_vec_withhalo("p", p);
                }

                /* v = Ap */
/*15*/          mult_A_vec(p, v);
                if ((print_level>=5) || ( (print_level>=4)&&(iter==0) )) print_vec("v", v);

                /* lambda = alpha / (v, p)2 */
/*15*/          lambda = alpha / dot_product_vec_vec(v, p);
                if ((print_level>=5) || ( (print_level>=4)&&(iter==0) )) {
                        printf("CG - my_rank =%3d, i=%d, vp=%9.6f, lambda=%9.6f\n",
                                my_rank,iter, dot_product_vec_vec(v, p), lambda); fflush(stdout);
                }

                /* x := x + lambda * p */
/*15*/          add_vec_vec(x, lambda, p, x);
                if ((print_level>=5) || ( (print_level>=4)&&(iter==0) )) print_vec("x", x);

                /* r := r - lambda * v */
/*15*/          add_vec_vec(r, -lambda, v, r);
                if ((print_level>=5) || ( (print_level>=4)&&(iter==0) )) print_vec("r", r);

                /* alpha_new = ||r||**2,   Caution: sqr_norm_vec() computes already ||.||**2 */
/*15*/          alpha_new = sqr_norm_vec(r);
                if ((print_level>=5) || ( (print_level>=4)&&(iter==0) )) {
                        printf("CG - my_rank =%3d, i=%d, alpha_new=%9.6f\n",
                                my_rank,iter, alpha_new); fflush(stdout);
                }

                /* p := r + (alpha_new/alpha) * p */
/*15*/          add_vec_vec(r, alpha_new/alpha, p, p);
                if ((print_level>=5) || ( (print_level>=4)&&(iter==0) )) print_vec("p", p);

                /* alpha := alpha_new */
/*15*/          alpha = alpha_new;
                if ((print_level>=2) && (my_rank==0))
                     printf("CG i=%d, ||r||**2 = %8.2e %s %8.2e,   ||r|| = %8.2e %s %8.2e\n",iter,
                                alpha_new,(alpha_new<=epsilon ? "<=" : "> "),epsilon,
                                sqrt(alpha_new),(alpha_new<=epsilon ? "<=" : "> "),sqrt(epsilon));
        }
        /* update output-arguments */ 
/*15*/  *num_iter = iter;
/*15*/  *norm_r = sqrt(alpha);
}

/*>>>>>>>>>>>>>>>>>>>>>> -end- of task 15 <<<<<<<<<<<<<<<<<<<<<<*/
/*>>>>>>>>>>>>>>>>>>>>>> begin of task 16 <<<<<<<<<<<<<<<<<<<<<<*/

/************************ main program *************************/
int main(int argc, char *argv[])
{
        int num_iter;
        double norm_r, norm_err, maxnorm_err;
        vector_chunk u, e;
        double time_start=0, time_end=0;

#ifdef serial
        numprocs = 1; my_rank = 0;
#else
        MPI_Init(&argc, &argv);
        MPI_Comm_size(MPI_COMM_WORLD, &numprocs);
        MPI_Comm_rank(MPI_COMM_WORLD, &my_rank);
#endif

        m = 4;
        n = 4;
        iter_max = 500;
        epsilon = 1e-6;
        print_level = 1;
        decomp_dims = 1;
        mprt = 11; nprt = 11;
        m_procs=2; n_procs=2;
#       ifndef serial
        {   int dims[2]; dims[0] = 0; dims[1] = 0;
            MPI_Dims_create(numprocs, 2, dims);
            m_procs = dims[0]; n_procs = dims[1];
        }
#       endif


        /* Parse command line options */
        read_options( argc, argv );

        Init_Decomp();
        if (print_level>=4) Print_Decomp();

        /* Init the distribution of the matrix and the vectors */
        Distribution(m*n);

        /* reserve chunk of matrix A */
        alloc_A();
        Init_Laplace_A(&A);
        if (print_level>=4) print_mat("A", &A);

#ifndef serial
        /* !!! determine communication partners for vector x !!! */
        Init_Vec_Halo();
        if (print_level>=4) print_halo();

        Init_Recv_Info();
        if (print_level>=4) print_recv_info();

        Init_Send_Info1();
        Init_Send_Info2();
        if (print_level>=4) print_send_info();
#endif

        /* reserve chunk of vector b and !!! assign border data */
        b = alloc_vec(0);
        Init_Boundary_Vec(b);

        if (print_level>=4) print_vec("b",b);

        /* x is initialized with zeros */
        x = alloc_vec(0);

#ifndef serial
        time_start = MPI_Wtime();
#endif

        /* solves Ax = b,  A,x,b are global variables */
        CG_solve(&num_iter, &norm_r);

#ifndef serial
        time_end = MPI_Wtime();
#endif

        u = alloc_vec(0);
        e = alloc_vec(0);
        Init_Exact_Solution_vec(u);
        /* p = u - x */
        add_vec_vec(u, -1, x, e);
        norm_err = sqrt(sqr_norm_vec(e));
        maxnorm_err = max_norm_vec(e);

        if (print_level>=3) print_vec2d("x",x,m,n);

        if (my_rank==0)
        {
            printf("main: %3d iterations, sqrt(eps) = %8.2e,  normalized: %8.2e\n",num_iter,
                                             sqrt(epsilon), sqrt(epsilon/(n*m)));
            printf("main:                 ||r||     = %8.2e,  normalized: %8.2e\n",
                                             norm_r, norm_r/sqrt(n*m));
            printf("main:                 ||error|| = %8.2e,  normalized: %8.2e\n",
                                             norm_err, norm_err/sqrt(n*m));
            printf("main:                 max(error)= %8.2e = normalized: %8.2e\n",
                                             maxnorm_err, maxnorm_err);
            printf("main:                 n*m = %d*%d = %d\n",n, m, n*m);
            printf("main:                 %8.2e sec\n",time_end-time_start);
        }

        if (print_level>=1) print_vec2d("x",x,mprt,nprt);

#ifndef serial
        MPI_Finalize();
#endif
        return 0;

#ifndef serial
        /* vorzeitiger Test-Stop */ MPI_Finalize();
#endif
        /* vorzeitiger Test-Stop */ return 0;
}
/*>>>>>>>>>>>>>>>>>>>>>> -end- of task 16 <<<<<<<<<<<<<<<<<<<<<<*/
