/* CG/solution/parts/cgp00.c */

/****************************************************************
 *                                                              *
 * This file has been written as a sample solution to an        *
 * exercise in a course given at the High Performance           *
 * Computing Centre Stuttgart (HLRS).                           *
 * It is made freely available with the understanding that      *
 * every copy of this file must include this header and that    *
 * HLRS takes no responsibility for the use of the enclosed     *
 * teaching material.                                           *
 *                                                              *
 * This software is partitioned into 16 tasks.                  *
 * Each task can be implemented by one group of the course      *
 * Files: CG/solution/cg_all.c - Solution                       *
 *        CG/skel/cg_01.c      - Skeletons that should be used  *
 *             .. cg_15.c        to implement each task         *
 *        CG/solution/cgp00.c  - Source part of each task       *
 *             ..     cgp16.c                                   *
 *                                                              *
 * Authors: Dr. Rolf Rabenseifner                               *
 *          Felix Triebel                                       *
 *          Michael Speck                                       *
 *          Felix Triebel                                       *
 *          Traugott Streicher                                  *
 *                                                              *
 * Contact: rabenseifner@hlrs.de                                *
 *                                                              *
 ****************************************************************/
                                                           /*EOH*/
/*>>>>>>>>>>>>>>>>>>>>>> begin of task 00 <<<<<<<<<<<<<<<<<<<<<<*/

#include <stdio.h>
#include <stdlib.h>
#ifndef serial
#  include <mpi.h>
#endif
#include <math.h>
#include <malloc.h>
#include <string.h>

/****************** declarations and global data ***************/

/* options: */
int     m,n;         /* size of the physical-heat matrix without borders */
                     /* m=vertical, n=horizontal */
int     iter_max;    /* maximal number of iterations in CG */
double  epsilon;     /* Limit for sqr( || residual vector || ) */
int     print_level; /* print_level = 1: Only result-error and partial result matrix */
                                 /* = 2: and residual norm after each iteration */
                                 /* = 3: and result physical-heat matrix */
                                 /* = 4: and all vector and matrix data in first iteration */
                                 /* = 5: and all vector and matrix data in _all_ iterations */
int     decomp_dims; /* dimensins of the domain decomposition */
int     m_procs, n_procs; /* number of processes in both directions of the 2-dim domain decomposition */
int     mprt, nprt;  /* number of rows and columns when printing the partial result matrix */

/* constants: */

#define MSG_TAG  0
#define MSG_TAG1 1
#define MSG_TAG2 2

#define WITH_HALO    1
#define WITHOUT_HALO 0

/* global structures and data: */

int numprocs, my_rank;

/* start and end+1 of the local chunk */
int start, end1;
int chunk_size;

 
/* data struktures */
struct matrix_row {
        int     size;
        int*    J; /* position in global row */
        int*    J_loc; /* position in local vectors */
        double* value;
};

/* chunk of the A matrix (each process gets one chunk) */
struct matrix_chunk {
        /* size = end1 - start */
        struct matrix_row* row;
} A;


 
/* chunk of a vector */
/* size = end1 - start + halo_size, if halo is used
 *                     + 0        , if halo is not used */
typedef double* vector_chunk;
vector_chunk x, b;

/* halo strukture, etc. */
struct {
    int  size;           /* number of already used elements */
    int  size_allocated; /* number of allocated elements in memory ( >=size ) */
    int* I;              /* used size = halo_info.size, index = halo_pos */
} halo_info;

struct neighbor_struct {
        int     size;    // number of elements to communicate
        int*    I_loc;   // positions in the halo or the real data
        double* scratch; // contiguous scratch array
};

/* neighbors from which we receive data into the halo */
struct neighbor_struct* recv_halo;  // Array of size numprocs
                     // recv_halo[rank].I_loc: positions in the halo, to store the values
int                     recv_num;
int*                    recv_ranks; // Array of size recv_num

/* neighbors which we send original data to */
struct neighbor_struct* send_halo;  // Array of size numprocs
                     // send_halo[rank].I_loc: positions in the real data, to fetch the values
int                     send_num;
int*                    send_ranks; // Array of size recv_num



/********************** memory management **********************/

void abrt(void)
{
#ifndef serial
        MPI_Abort(MPI_COMM_WORLD, -1);
#endif
        exit(-1);
}

/* save_malloc: Same meaning as malloc(); error handling included */
void* save_malloc(int size)
{
        void *new;

        if (size == 0) return NULL;

        new = (void*)malloc(size);
        if (new == NULL) {
                printf("Could not allocate memory.\n");
                abrt();
        } 
        return(new);

}

/*------------------------------
 * This helper function is needed for dynamically growing integer arrays.
 * It takes a pointer to the array and a pointer to a size_allocated variable
 * as arguments.
 *
 * average memory wastage: 33% (in case of normal distribution of size),
 *    worst case 50%
 * average time used for (re)initializing data: 2*sizeof(array_used)
 *    = 1.5 * 2 * sizeof(array_needed) in average case
 *    or 2 * 2 * sizeof(array_needed) in worst case.
 * We use this solution, as it is only needed twice:
 *    for initializing halo and communication info
 *    before computation.
 */
void enlarge_int_array(int** array, int* size_allocated)
{
        /* size_allocated holds the old size of the array and gets altered here. */
        *size_allocated = 2 * (*size_allocated) + 1;  // +1 if size_allocated = 0
        *array = (int*)realloc(*array, sizeof(int) * (*size_allocated));
        if (*array == NULL) {
                printf("Could not allocate memory.\n");
                abrt();
        }
}


/********************** commandline options ********************/

void read_options( int argc, char* argv[] )
{
        int optid = 1; /* start at first optional argument */
        int err_detected = 0;
        if ( my_rank == 0 ) 
        {        
            while ( optid < argc ) {
                if (!strcmp( argv[optid], "-m") ) { optid++;
                    m        = atoi( argv[optid] ); if (m<=0) {err_detected = 1; break;}
                } else if (!strcmp( argv[optid], "-n") ) { optid++;
                    n        = atoi( argv[optid] ); if (n<=0) {err_detected = 1; break;}
                } else if (!strcmp( argv[optid], "-imax") ) { optid++;
                    iter_max = atoi( argv[optid] ); if (iter_max<=0) {err_detected = 1; break;}
                } else if (!strcmp( argv[optid], "-eps") ) { optid++;
                    epsilon  = atof( argv[optid] ); if (epsilon<=0) {err_detected = 1; break;}
                } else if (!strcmp( argv[optid], "-prtlev") ) { optid++;
                    print_level = atoi( argv[optid] ); if (print_level<0) {err_detected = 1; break;}
                } else if (!strcmp( argv[optid], "-twodims") ) { 
                    decomp_dims = 2;
                } else if (!strcmp( argv[optid], "-mprocs") ) { optid++;
                    m_procs = atoi( argv[optid] ); if (m_procs<=0) {err_detected = 1; break;}
                } else if (!strcmp( argv[optid], "-nprocs") ) { optid++;
                    n_procs = atoi( argv[optid] ); if (n_procs<=0) {err_detected = 1; break;}
                } else {err_detected = 1; break;}
                optid++;
            }
            if ( err_detected ) {
                printf( "unknown command line option: %s\n", argv[optid] );
                printf( "\n");
                printf( "Synopsis: mpirun -np <number of processes> ./cg_all <options>\n");
                printf( "  with options [defaults]:\n");
                printf( "  -m <number of rows in physical area of the Laplace equation> [4]\n");
                printf( "  -n <number of columns in physical area of the Laplace equation [4]\n");
                printf( "  -imax <maximal number of iterations> [500]\n");
                printf( "  -eps  <epsion> [1e-6]\n");
                printf( "  -prtlev <print and debug level 0..5> [1]\n");
                printf( "  -twodims [only one-dimensional domain decomposition (DD)]\n");
                printf( "  -mprocs <horizontal number of processes in 2-dim DD\n");
                printf( "  -nprocs <vertical   number of processes in 2-dim DD\n");
                printf( "\n");
                fflush(stdout); abrt();
            }
        }
#ifndef serial
        MPI_Bcast(&m,          1, MPI_INT,    0, MPI_COMM_WORLD);
        MPI_Bcast(&n,          1, MPI_INT,    0, MPI_COMM_WORLD);
        MPI_Bcast(&iter_max,   1, MPI_INT,    0, MPI_COMM_WORLD);
        MPI_Bcast(&epsilon,    1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
        MPI_Bcast(&print_level,1, MPI_INT,    0, MPI_COMM_WORLD);
        MPI_Bcast(&decomp_dims,1, MPI_INT,    0, MPI_COMM_WORLD);
        MPI_Bcast(&m_procs,    1, MPI_INT,    0, MPI_COMM_WORLD);
        MPI_Bcast(&n_procs,    1, MPI_INT,    0, MPI_COMM_WORLD);
#endif
}

/*>>>>>>>>>>>>>>>>>>>>>> -end- of task 00 <<<<<<<<<<<<<<<<<<<<<<*/
