/* CG/solution/parts/cgp15.c */

/****************************************************************
 *                                                              *
 * This file has been written as a sample solution to an        *
 * exercise in a course given at the High Performance           *
 * Computing Centre Stuttgart (HLRS).                           *
 * It is made freely available with the understanding that      *
 * every copy of this file must include this header and that    *
 * HLRS takes no responsibility for the use of the enclosed     *
 * teaching material.                                           *
 *                                                              *
 * This software is partitioned into 16 tasks.                  *
 * Each task can be implemented by one group of the course      *
 * Files: CG/solution/cg_all.c - Solution                       *
 *        CG/skel/cg_01.c      - Skeletons that should be used  *
 *             .. cg_15.c        to implement each task         *
 *        CG/solution/cgp00.c  - Source part of each task       *
 *             ..     cgp16.c                                   *
 *                                                              *
 * Authors: Dr. Rolf Rabenseifner                               *
 *          Felix Triebel                                       *
 *          Michael Speck                                       *
 *          Felix Triebel                                       *
 *          Traugott Streicher                                  *
 *                                                              *
 * Contact: rabenseifner@hlrs.de                                *
 *                                                              *
 ****************************************************************/
                                                           /*EOH*/
/*>>>>>>>>>>>>>>>>>>>>>> begin of task 15 <<<<<<<<<<<<<<<<<<<<<<*/

/************************ CG solver ****************************/

/*------------------------------
 * void CG_solve
 * OUT int *num_iter  : return the number of iterations
 * OUT double *norm_r : return the norm of r
 *
 * TASK: solve Laplace equation
 *
 * REFERENCE: Slide 12, 24
 *
 * GLOBAL DATA:
 * IN int iter_max       : maximal number of iterations
 * IN double epsilon     : epsilon for maximal error
 * IN struct matrix_chunk A: Rows contain columns of elements we need.
 * IN vector_chunk b     : local chunk of vector b
 * IN/OUT vector_chunk x : local chunk of vector x
 * USED FUNCTIONS: APPL: alloc_vec, dupl_vec, comm_vec_halo, mult_A_vec,
 *                       add_vec_vec, sqr_norm_vec, dot_product_vec_vec
 */
void CG_solve(int *num_iter, double *norm_r)
{
        /* only p has a halo */
        vector_chunk p, r, v;
        double alpha, alpha_new, lambda;
        int iter;

        /* allocate vectors p, r, v */ 
        p = alloc_vec(1);
        r = alloc_vec(0);
        v = alloc_vec(0);

        /* x is initialized with 0 */

        /* instead of p = Ax */
           /* p := x */
           dupl_vec(x, p);
#ifndef serial
           /* update halo */
           comm_vec_halo(p);
#endif

           /* v = Ap */
           mult_A_vec(p, v);
           if (print_level>=4) print_vec("v", v);

           /* p = b - v */
           add_vec_vec(b, -1, v, p);
           if (print_level>=4) print_vec("p", p);

        /* r := p */
        dupl_vec(p, r);
        if (print_level>=4) print_vec("r", r);

        /* alpha = ||r||**2,   Caution: sqr_norm_vec() computes already ||.||**2 */
        alpha = sqr_norm_vec(r);


        for(iter=0; (iter < iter_max) && (alpha > epsilon) ; iter++)
        {
                if ((print_level>=5) || ( (print_level>=4)&&(iter==0) )) {
                printf("CG - my_rank =%3d, i=%d, alpha=%9.6f\n",my_rank,iter,alpha); fflush(stdout);
                }

#ifndef serial
                /* update halo */
                comm_vec_halo(p);
#endif

                if ((print_level>=5) || ( (print_level>=4)&&(iter==0) )) {
                        print_mat("A", &A);
                        print_vec("p", p);
                        print_vec_withhalo("p", p);
                }

                /* v = Ap */
                mult_A_vec(p, v);
                if ((print_level>=5) || ( (print_level>=4)&&(iter==0) )) print_vec("v", v);

                /* lambda = alpha / (v, p)2 */
                lambda = alpha / dot_product_vec_vec(v, p);
                if ((print_level>=5) || ( (print_level>=4)&&(iter==0) )) {
                        printf("CG - my_rank =%3d, i=%d, vp=%9.6f, lambda=%9.6f\n",
                                my_rank,iter, dot_product_vec_vec(v, p), lambda); fflush(stdout);
                }

                /* x := x + lambda * p */
                add_vec_vec(x, lambda, p, x);
                if ((print_level>=5) || ( (print_level>=4)&&(iter==0) )) print_vec("x", x);

                /* r := r - lambda * v */
                add_vec_vec(r, -lambda, v, r);
                if ((print_level>=5) || ( (print_level>=4)&&(iter==0) )) print_vec("r", r);

                /* alpha_new = ||r||**2,   Caution: sqr_norm_vec() computes already ||.||**2 */
                alpha_new = sqr_norm_vec(r);
                if ((print_level>=5) || ( (print_level>=4)&&(iter==0) )) {
                        printf("CG - my_rank =%3d, i=%d, alpha_new=%9.6f\n",
                                my_rank,iter, alpha_new); fflush(stdout);
                }

                /* p := r + (alpha_new/alpha) * p */
                add_vec_vec(r, alpha_new/alpha, p, p);
                if ((print_level>=5) || ( (print_level>=4)&&(iter==0) )) print_vec("p", p);

                /* alpha := alpha_new */
                alpha = alpha_new;
                if ((print_level>=2) && (my_rank==0))
                     printf("CG i=%d, ||r||**2 = %8.2e %s %8.2e,   ||r|| = %8.2e %s %8.2e\n",iter,
                                alpha_new,(alpha_new<=epsilon ? "<=" : "> "),epsilon,
                                sqrt(alpha_new),(alpha_new<=epsilon ? "<=" : "> "),sqrt(epsilon));
        }
        /* update output-arguments */ 
        *num_iter = iter;
        *norm_r = sqrt(alpha);
}

/*>>>>>>>>>>>>>>>>>>>>>> -end- of task 15 <<<<<<<<<<<<<<<<<<<<<<*/
